#include <stdarg.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/timeb.h>

#include "config.h"
#include "utils.h"
#include "wildmat.h"
#include "rfcutils.h"
#include "pbbs_out.h"
#include "log.h"

struct stats {
	long int messages, personals, bulletins, smtp, nntp;
	long int bounces, errors;
} st;

char monstr[][4] =	{ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", 
			"Sep", "Oct", "Nov", "Dec" };
char daystr[][4] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

struct newsgroup_map	*pbbsl		= NULL;

FILE *bf;
FILE *f;
char bfname[PATHLEN];
int msgc = 0, linec = 0;

void *stripnewline (char *i); 
int find_more_Tlines (char *bodycarry);

/*
 *	Crash, log the reason
 */

void panic(const char *fmt, ...)
{
	va_list args;
	
	va_start(args, fmt);
	say(L_ERR, fmt, args);
	va_end(args);

	exit(1);
}

/*
 *	Parse a PBBS style date/time string: yymmdd/hhmmZ
 */

time_t parsetime(char *s)
{
	struct tm t = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	char ss[4];
	
	/* FIXME: This WILL SCREW UP BAD if the date/time string is,
	   for example, shorter than expected, or otherwise corrupted */
	ss[0] = *s++; ss[1] = *s++; ss[2] = '\0';
	if (ss[0] == '0') ss[0] = ' ';
	sscanf(ss, "%i", &t.tm_year);
	ss[0] = *s++; ss[1] = *s++;
	if (ss[0] == '0') ss[0] = ' ';
	sscanf(ss, "%i", &t.tm_mon);
	ss[0] = *s++; ss[1] = *s++;
	sscanf(ss, "%i", &t.tm_mday);
	s++; /* skip / */
	ss[0] = *s++; ss[1] = *s++;
	sscanf(ss, "%i", &t.tm_hour);
	ss[0] = *s++; ss[1] = *s++;
	sscanf(ss, "%i", &t.tm_min);
	
	t.tm_mon--;
	if (t.tm_year < 34) /* Y2K workaround */
		t.tm_year += 100;
	
	return mktime(&t);
}

/*
 *	Create a NNTP/SMTP style date/time string
 */
 
char *datestr(time_t t)
{
	static char s[40];
	time_t ti = t;
	struct tm *tm;
	
	tm = gmtime(&ti);
	snprintf(s, 40, "%s, %i %s %i %02i:%02i:%02i +0000",
		daystr[tm->tm_wday], tm->tm_mday, monstr[tm->tm_mon],
		tm->tm_year + 1900, tm->tm_hour, tm->tm_min, tm->tm_sec);
	s[39] = '\0'; /* Hey, just in case... */
	return s;
}

/*
 *	Create a NNTP/SMTP style MID from a pbbs [MB]ID
 */
 
char *bidtomid(char *bid)
{
	char mid[40];
	char *s;
	
	snprintf(mid, 40, "<%s@hamradio>", bid);
	mid[39] = '\0';
	if (!(s = strdup(mid)))
		panic("Out of memory!");
	return s;
}

/*
 *	Read a line, removing any creeping DOS newlines 
 */

char *readl(FILE *f)
{
	static char s[LINELEN];
	char *p;
	
	if (!(fgets((char *)&s, LINELEN, f)))
		return NULL; /* EOF */
	
	for (p = s; (*p); p++)
		if (*p == '\r') {
			*p = '\n';
			p++;
			*p = '\0';
		} else if (*p == '\n') {
			p++;
			*p = '\0';
		}
	
	return s;
}

/*
 *	Read a line from *f using readl(), increasing line counter
 */
 
char *readln(void)
{
	linec++;
	return readl(f);
}

/*
 *	Pick newsgroup for the BBS distribution
 */

struct newsgroup_map *pick_newsgroup(char *to, char *tobbs, char *from, char *frombbs)
{
	struct newsgroup_map *m;

	for (m = pbbsl; (m); m = m->next) {
		if (((!m->to) || wildmat(to, m->to, 0)) 
			&& ((!m->bbs) || wildmat(tobbs, m->bbs, 0))
			&& ((!m->from) || wildmat(from, m->from, 0))
			&& ((!m->frombbs) || wildmat(frombbs, m->frombbs, 0)))
				return m;
	}

	return NULL;
}

/*
 *	Read and handle a message
 */

int handlemesg(void)
{
	char *s, *t;
	int argc, arg = 2, foundt = 0;
	char *argv[32];
	
	struct pb_hdr pb = { "", 0, "", "", "", "", "", "", "", 0 };
	char pbsrc[CALLEN + HRLEN], pbdst[CALLEN + HRLEN];
	char pbpath[LINELEN] = "", *dist;
	char *src = NULL, *dst = NULL, *newsgroups = NULL, *mid = NULL;
        char bodycarry [LINELEN];
	
	struct timeb tp;
	
	char tfname[PATHLEN], str[PATHLEN];
	FILE *tf;
	
	struct newsgroup_map *m;
	
/* Read and parse message header */
	if (!(s = readln()))
		return 0; /* EOF */
	
	if ((argc = parse_args(argv, s)) == 0) {
		say(L_ERR, "Unexpected empty line (line %i msg %i)", 
			linec, msgc);
		return -1;
	}
	
	if (argc < 5) {
		say(L_ERR, "Not enough arguments (%i) (line %i msg %i)", 
			argc, linec, msgc);
		return -2;
	}
	
	if ((!strcasecmp(argv[0], "SP")) || (!strcasecmp(argv[0], "SA"))) {
		pb.mtype = MT_P;
		st.personals++;
	} else
	if (!strcasecmp(argv[0], "SB")) {
		pb.mtype = MT_B;
		st.bulletins++;
	} else {
		say(L_ERR, "Unrecognized S command: %s (line %i msg %i)", 
			argv[0], linec, msgc);
		return -3;
	}
	
	hstrncpy(pb.to, argv[1], CALLEN);
	hstrncpy(pb.tobbs, hroute, HRLEN); /* Just in case there's no @ field */
	
	while (arg < argc) {
		if (*argv[arg] == '@') {
			hstrncpy(pb.tobbs, argv[arg+1], HRLEN);
		} else if (*argv[arg] == '<') {
			hstrncpy(pb.from, argv[arg+1], CALLEN);
		} else if (*argv[arg] == '$') {
			hstrncpy(pb.bid, &argv[arg][1], BIDLEN);
		}
		arg = arg + 2;
	}
	
	if (!pb.bid[0]) {
		say(L_ERR, "No BID: %s > %s@%s (msg %i)", 
			pb.from, pb.to, pb.tobbs, msgc);
		return -3;
	} else if (!pb.from[0]) {
		say(L_ERR, "$%s: No source address on S line (line %i msg %i)", 
			pb.bid, linec, msgc);
		return -3;
	}
	
	/* Read message subject */
	if (!(s = readln())) {
		say(L_ERR, "$%s: Unexpected EOF at subject (line %i msg %i)", 
			pb.bid, linec, msgc);
		return -4; /* EOF */
	}
	
	hstrncpy(pb.subj, s, LINELEN);
	
/* Read the R: lines */
	while (1) {
		if (!(s = readln())) {
			say(L_ERR, "$%s: Unexpected EOF at R: lines (line %i msg %i)", 
				pb.bid, linec, msgc);
			return -5; /* EOF */
		}

	        if (!strncmp(s, "T:", 2))
			break; /* Before the T: line is broken by parse_args */
	   
		if ((argc = parse_args(argv, s)) < 2)
			break; /* Expected empty line */
		
		if (strncmp(s, "R:", 2))
			break; /* Finished with R: lines */
		
		pb.time = parsetime(&argv[0][2]);
		if (!strncmp(argv[1], "@:", 2))
			hstrncpy(pb.frombbs, &argv[1][2], HRLEN);
		else if (!strncmp(argv[1], "@", 1))
			hstrncpy(pb.frombbs, &argv[1][2], HRLEN);
		else if (strchr(argv[1], '@')) {
			strtok(argv[1], "@");
			hstrncpy(pb.frombbs, strtok(NULL, "@"), HRLEN);
			if (pb.frombbs[0] == ':')
				memmove(&pb.frombbs[0], &pb.frombbs[1], HRLEN-1);
		} else
			hstrncpy(pb.frombbs, argv[1], HRLEN); /* last resort */
		
		strncat(pbpath, pb.frombbs, LINELEN);
		strncat(pbpath, "!", LINELEN);
	}
	
/* Read the T: lines */

	while (1) {
	        if (strncmp(s, "T:", 2)) {
		        if (foundt)
			        break;
		   
		        if ((foundt = find_more_Tlines (bodycarry)) == 0)
			        break;
		        
			else if (!(s = readln()))
			        break; /* EOF */		   
		}
		   
		foundt = 1;

/* G4APL hates T:From lines */
#if 0	  
	        if ((!strncasecmp(s, "T:From: ", 8)) && (src == NULL)) {
			src = strdup(&s[8]);
		        stripnewline (src);
		} else
#endif
		if ((!strncasecmp(s, "T:To: ", 6)) && (dst == NULL)) {
			dst = strdup(&s[6]);
		        stripnewline (dst);
		} else
		if ((!strncasecmp(s, "T:Newsgroups: ", 14)) && (newsgroups == NULL)) {
			newsgroups = strdup(&s[14]);
		   	stripnewline (newsgroups);
		} else
		if ((!strncasecmp(s, "T:Message-ID: ", 14)) && (mid == NULL)) {
			mid = strdup(&s[14]);
		        stripnewline (mid);
		}
		
		if (!(s = readln()))
			break; /* EOF */
	}
	
/* Decide the routing (smtp or nntp), and do the rewriting. */
	snprintf(pbsrc, CALLEN + HRLEN, "%s@%s", pb.from, pb.frombbs);
	snprintf(pbdst, CALLEN + HRLEN, "%s@%s", pb.to, pb.tobbs);
	
	/* Rewrite source address, if it wasn't found on T: lines */
	if (src) {
		/* we have it from the T: line already. */
	} else if ((src = map_address(pbsrc, PBBS_SRC))) {
		t = strdup(src);
		src = t;
		strlwr(src);
	} else {
		say(L_R_ERR, "$%s %s > %s: No source mapping",
			pb.bid, pbsrc, pbdst);
		pb.mtype = MT_NONE;
	}
	ftime(&tp);
	
	if (!mid)
		mid = bidtomid(pb.bid);
	
	switch (pb.mtype) {
	
	case MT_P:	/* Personal message => smtp */
		
		/* Rewrite destination address, if necessary */
		if (dst) {
			/* we have it from the T: line already. */
		} else if ((dst = map_address(pbdst, PBBS_DST))) {
			t = strdup(dst);
			dst = t;
			strlwr(dst);
		} else {
			say(L_R_ERR, "$%s %s > %s: No dest mapping",
				pb.bid, pbsrc, pbdst);
			pb.mtype = MT_NONE;
			break;
		}
		
		say(L_ROUTING, "$%s %s > %s => %s %s > %s", 
			pb.bid, pbsrc, pbdst, mid, src, dst);
		
		snprintf(tfname, PATHLEN, "/tmp/%i.%i.%i.tmp", tp.time, tp.millitm, msgc);
		if (!(tf = fopen(tfname, "w"))) {
			say(L_ERR, "$%s: Cannot open temporary file %s: %s. Message lost.", 
				pb.bid, tfname, strerror(errno));
			pb.mtype = MT_NONE;
		}
		
		fprintf(tf, "Received: from %s (Amateur Packet Radio BBS)\n", hroute);
		fprintf(tf, "\tby %s (mailgw %s) with PBBS id %s\n", hostname, VERSION, pb.bid);
		fprintf(tf, "\tfor <%s@%s>; %s\n", pb.to, pb.tobbs, datestr(time(NULL)) );
		fprintf(tf, "Date: %s\n", datestr(pb.time));
		fprintf(tf, "From: %s\n", src);
		fprintf(tf, "To: %s\n", dst);
		fprintf(tf, "Subject: %s", pb.subj);
		fprintf(tf, "Message-ID: %s\n", mid);
		fprintf(tf, "X-BID: %s\n", pb.bid);
		fprintf(tf, "X-BBS-From: %s@%s\n", pb.from, pb.frombbs);
		fprintf(tf, "X-BBS-To: %s@%s\n", pb.to, pb.tobbs);
		fprintf(tf, "X-BBS-Path: %s\n", pbpath);
		fprintf(tf, "Return-Path: <%s>\n", src);
		fprintf(tf, "\n");
		
		break;

	case MT_B:	/* Bulletin => nntp */
		
		if (newsgroups) {
			/* we have it from the T: line already. */
		} else if ((m = pick_newsgroup(pb.to, pb.tobbs, pb.from, pb.frombbs))) {
			newsgroups = strdup(m->newsgroup);
		} else {
			say(L_R_ERR, "$%s %s > %s: No newsgroup mapping",
				pb.bid, pbsrc, pbdst);
			pb.mtype = MT_NONE;
			break;
		}

		say(L_ROUTING, "$%s %s > %s => %s %s > %s",
			pb.bid, pbsrc, pbdst, mid, src, newsgroups);

		snprintf(tfname, PATHLEN, "%s/%i.%i.%i.tmp", spooldir, tp.time, tp.millitm, msgc);
	        if (!(tf = fopen(tfname, "w"))) {
			say(L_ERR, "Cannot open temporary file %s: %s. Message lost.", 
				tfname, strerror(errno));
			pb.mtype = MT_NONE;
		}
		
		fprintf(tf, "Path: %s!packet-news-gateway\n", hostname);
		fprintf(tf, "From: %s\n", src);
		fprintf(tf, "Newsgroups: %s\n", newsgroups);

	        if ((dist = map_distribution (pb.tobbs, PBBS2NNTP)))
	                fprintf (tf, "Distribution: %s\n", dist);
	   
		fprintf(tf, "Subject: %s", pb.subj);
		fprintf(tf, "Date: %s\n", datestr(pb.time));
		fprintf(tf, "Organization: Amateur Packet Radio Network\n");
		fprintf(tf, "Message-ID: %s\n", mid);
		fprintf(tf, "NNTP-Posting-Host: %s\n", hostname);
		fprintf(tf, "X-BID: %s\n", pb.bid);
		fprintf(tf, "X-BBS-From: %s@%s\n", pb.from, pb.frombbs);
		fprintf(tf, "X-BBS-To: %s@%s\n", pb.to, pb.tobbs);
		fprintf(tf, "X-BBS-Path: %s\n", pbpath);
		fprintf(tf, "\n");
		
		break;
	}
	
	free(src);
	free(dst);
	free(newsgroups);


/* Replace any From or To lines encountered when looking for T lines */
#if 1
	        if (*bodycarry != '\0') {
		        strcat (bodycarry, "\n");
	                fputs(bodycarry, tf);
		}
#endif	   
   
/* Read the message body */
	while (1) {
		if (!(s = readln()))
			break; /* EOF */
		
		if (!strcmp(s, "/EX\n"))
			break; /* Finished with message body */
		
		if (pb.mtype)
			fputs(s, tf);
	}

	if (pb.mtype) {
		fclose(tf);
		switch (pb.mtype) {
		
		case MT_B:
			fprintf(bf, "%s %s\n", tfname, mid);
			fflush(bf);
			st.nntp++;
			break;
		
		case MT_P:
			sprintf(str, "%s -oem -t < %s", sendmail, tfname);
			system(str);
			if (remove(tfname))
				say(L_ERR, "%s: Could not remove temporary file %s: %s",
					pb.bid, tfname, strerror(errno));
			st.smtp++;
			break;
		}
	}
	
	free(mid);
	
	return 1;
}

/*
 *	Process the export file
 */

int process(void)
{
	struct lhandle *lh;
	struct stat sbuf;
	int i = 0;
	
	if (stat(exportfile, &sbuf)) {
		if (errno == ENOENT) {
			say(L_INFO, "%s doesn't exist.", exportfile);
			return 0; /* Doesn't exist */
		}
		say(L_ERR, "Cannot stat() %s: %s!", exportfile, strerror(errno));
		return -1;
	}
	
	if (!(lh = fbblock(exportlock))) {
		say(L_INFO, "Lock %s present, giving up.", exportlock);
		return -1;
	}
	
	if (!(f = fopen(exportfile, "r"))) {
		say(L_ERR, "Cannot open %s: %s!", exportfile, strerror(errno));
		return -1;
	}
	
	snprintf(bfname, PATHLEN, "%s/innxmit.batch", spooldir);
	if (!(bf = fopen(bfname, "a"))) {
		say(L_ERR, "Cannot open %s: %s!", bfname, strerror(errno));
		return -1;
	}
	
	setenv("TZ", "UTC 0", 1);	/* Any nicer way to get mktime() */
	tzset();			/* and friends to talk UTC ?? */
	
	do {
		msgc++;
		i = handlemesg();
		if (i == 1)
			st.messages++;
		else if (i < 0) {
			st.errors++;
			printf("Error: %i\n", i);
		}
	} while (i != 0);
	
	fclose(f);
	if (remove(exportfile)) {
		say(L_ERR, "Couldn't remove export file %s: %s",
			exportfile, strerror(errno));
	}
	fbbunlock(lh);

	fclose(bf);
	
	return 0;
}

/*
 *	Main
 */

int main(void)
{
	openlog("pbbs_rcv", LOG_NDELAY, LOG_LOCAL6);
	read_config();
	read_newsgroup_map(BULLETIN_MAP, 1, &pbbsl);
	
	umask(077); /* We don't want our temp files to be world readable */
	setvbuf(stdout, NULL, _IONBF, 0); /* No buffering on stdout, please */
	
	process();
	
	if (st.messages)
		say(L_STATS, "Stats: %li p => %li smtp, %li b => %li nntp",
			st.personals, st.smtp, st.bulletins, st.nntp);
	
	if (st.errors)
		say(L_STATS, "Errors: %li errors", st.errors);
	
	closelog();
	
	return 0;
}

/*
 *  Terminate a line at the first newline character
 */

void *stripnewline (char *i)
{
char *p;
        p = strchr(i,'\n');
	if (p)
                *p = '\0';
        return 0;
}

/*
 * Find out if there are T: Lines located near the top of the message body
 */

int find_more_Tlines (char *bodycarry)
{
int blanks = 0, count = 0, argc;
char *s, *argv[32];
fpos_t fp, fq;

        *bodycarry = '\0';
        fgetpos (f, &fp);                 /* fpointer to use if no more T's */ 

        while (count < 5 && blanks < 2) {
	        count ++;
	        fgetpos(f, &fq);         /* fpointer to use if there are T's */
	   
		if (!(s = readln()))
			break; /* EOF */
	   
	        if (!strncasecmp(s, "From: ", 6)) {
		        strcat (bodycarry, s);
		        continue;
		}
	        else if
		(!strncasecmp(s, "To  : ", 6)) {
		        strcat (bodycarry, s);
		        continue;
		}
	   
		else if 
	        ((argc = parse_args(argv, s)) < 1) {
			blanks ++; /* Empty line */
		        continue ;
		 	}
	        else if
		(!strncasecmp(s, "T:",2)) {
		        if (fsetpos(f, &fq) == 0) /* fpointer looking at T line */
		                return 1;
		        else {
			        panic("Error setting file pointer");
			        return -1;
			}
		}
	        else
	                break;    /* Not T, From, To, or Blank */
	   	}  /* while */

        *bodycarry = '\0';
   
        if (fsetpos(f, &fp) == 0) /* Sets the fpointer as if nothing happened */
                return 0;
	else {
		panic("Error setting file pointer");
	        return -1;
	}
}
